#!/bin/bash
########################################################################
# build_range.sh
#
# Build range and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_03_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: $*\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: $prog <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       $prog -h|--help [<product-dir>]

  Build $package and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking $prog.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking $prog.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "$@"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="$arg"; break
  done
  if [ -z "$product_dir" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> $product_dir\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "$product_dir/.upsfiles" ] && [ -w "$product_dir" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "$product_dir\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "$product_dir" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: $product_dir\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="$product_dir/ssibuildshims/$ssibuildshims_version"
  [ -f "$ssibuildshims_topdir/bin/ssi_functions" ] && \
    source "$ssibuildshims_topdir/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "$product_dir" && \
    setup ssibuildshims $ssibuildshims_version -z "$product_dir" || \
    ssi_die "unable to find ssibuildshims $ssibuildshims_version in
$product_dir and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "$product_dir" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  $trace verify_in_filesystem -d "$@" "$pkgdir/include"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "$maketar" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=(-e lib)
    $trace is_built "$@" && \
      $trace make_basic_distribution "$@" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
thisdir=$(cd "${BASH_SOURCE%/*}" && pwd -P)
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. $thisdir/version_info.sh || \
  ssi_die "unable to load version info from $thisdir/version_info"
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "$@"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_CONFIG
  DEBUG_BUILD
  DEBUG_TEST
  DEBUG_INSTALL

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.

  WANT_TESTS

    Activate tests.

EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
$trace define_basics TRACK_PROGRESS NO_QUAL NO_DEBUG_SOURCE NO_FLAVOR NO_FQ_DIR || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
$trace fake_declare_product -x -- || exit
####################################

####################################
# Build-only dependencies. *C*

setup_git --optional
setup_cmake --latest || exit
setup_ninja --optional && (( want_ninja = 1 ))
if (( WANT_TESTS )) || (( WANT_EXAMPLES )); then
  # modern_cxx() will set cc, cxx and cxxstd if successful.
  $trace modern_cxx 11 || {
    ssi_warn "$package requires a modern compiler for tests or examples"
    unset WANT_TESTS WANT_EXAMPLES
  }
fi
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
$trace maybe_make_distribution --silent

ssi_info "building $package for $OS-$plat-$qualdir (flavor $pkgflvr)"

# Ensure directories: pkgdir, sourcedir, blddir.
$trace init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

$trace cd "$sourcedir" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
$trace prepare_source "${ssi_args[@]}" -P "$patchdir" \
  "$tardir/$pkgtarfile" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "$DEBUG_PATCHES" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a cmake_options
(( want_ninja )) && cmake_options+=(-GNinja)

range_options_off=(
  RANGES_DEEP_STL_INTEGRATION
  RANGE_V3_DOCS
)

range_options_on=(RANGES_PREFER_REAL_CONCEPTS)

if (( WANT_EXAMPLES )); then
  ssi_info "examples built in $blddir/example will *not* be installed"
  range_options_on+=(
    RANGE_V3_EXAMPLES
    RANGES_BUILD_CALENDAR_EXAMPLE
  )
else
  range_options_off+=(RANGE_V3_EXAMPLES)
fi

if (( WANT_TESTS )); then
  range_options_on+=(
    RANGE_V3_HEADER_CHECKS
    RANGE_V3_TESTS
  )
else
  range_options_off+=(
    RANGE_V3_HEADER_CHECKS
    RANGE_V3_TESTS
  )
fi

if (( DEBUG_CONFIG )); then
  range_options_on+=(RANGES_VERBOSE_BUILD)
fi

cmake_options+=(
  -DCMAKE_BUILD_TYPE=Release
  -DCMAKE_INSTALL_PREFIX="$pkgdir"
  ${cxxstd:+-DRANGES_CXX_STD=$cxxstd}
)

for opt in ${range_options_on[*]:+"${range_options_on[@]}"}; do
  cmake_options+=(-D$opt:BOOL=ON)
done
for opt in ${range_options_off[*]:+"${range_options_off[@]}"}; do
  cmake_options+=(-D$opt:BOOL=OFF)
done

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "$CMAKE_BUILD_PARALLEL_LEVEL" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=$ncores

# For ctest:
[[ -n "$CTEST_PARALLEL_LEVEL" ]] || \
  export CTEST_PARALLEL_LEVEL=$ncores
##################
####################################

####################################
# Configure, build, and install.
#
# Use $blddir/$pkgtarname for in-source builds, $blddir otherwise.

$trace cd "$blddir" || \
  ssi_die "unable to cd to build directory $blddir"

##################
# Configure. *C*
$trace env CC=${cc} CXX=${cxx} \
       cmake ${cmake_options[*]:+"${cmake_options[@]}"} \
       "$sourcedir/$pkgtarname" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*
$trace cmake --build "$blddir" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
if (( WANT_TESTS )); then
  declare -a ctest_options
  if [ -n "$DEBUG_TEST" ]; then
    ctest_options+=(-V)
  else
    ctest_options+=(--output-on-failure)
  fi
  $trace ctest ${ctest_options[*]:+"${ctest_options[@]}"} || \
    ssi_die "tests failed"
fi
##################

##################
# Install. *C*

$trace cmake --install "$blddir" ${DEBUG_INSTALL:+--verbose} || \
  ssi_die "install failed"
##################

ssi_info "finished building $package $pkgver"

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
$trace declare_product -x -- || exit
##################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
$trace is_built && $trace ssi_build_completed && \
  ssi_info "$package is installed at $pkgdir" && \
  $trace maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
